@@ -52,7 +52,7 @@
 //!             Default::default())
 //!             .await?;
 //!
-//!         producer.produce(vec![1,2,3]).await?;
+//!         producer.produce("test message").await?;
 //!         Ok::<_, avassa_client::Error>(())
 //!     });
 //!
@@ -62,20 +62,19 @@
 //!         Default::default())
 //!         .await?;
 //!
-//!     let (_metadata, message) = consumer.consume().await?;
+//!     let message = consumer.consume().await?;
 //!
-//!     assert_eq!(message, vec![1,2,3]);
+//!     assert_eq!(message.payload, "test message");
 //!     Ok(())
 //! }
 //! ```
 
 #![deny(missing_docs)]
 
-use log::{debug, error};
 use serde::Deserialize;
 use serde_json::json;
 
-// pub mod strongbox;
+pub mod strongbox;
 pub mod volga;
 
 /// Description of an error from the REST APIs
@@ -110,7 +109,7 @@ pub enum Error {
 
     /// Websocket error
     #[error("Websocket error: {0}")]
-    WebSocket(#[from] tungstenite::Error),
+    WebSocket(#[from] tokio_tungstenite::tungstenite::Error),
 
     /// JSON serialization/deserialization error
     #[error("Serde JSON error: {0}")]
@@ -144,18 +143,39 @@ pub enum Error {
     /// IO Errors
     #[error("IO error {0}")]
     IO(#[from] std::io::Error),
+
+    /// General Error
+    #[error("Error {0}")]
+    General(String),
+}
+
+impl Error {
+    /// Create a general error
+    pub fn general(err: &str) -> Self {
+        Self::General(err.to_string())
+    }
 }
 
 /// Result type
 pub type Result<T> = std::result::Result<T, Error>;
 
-#[derive(Debug, Deserialize)]
+#[derive(Deserialize)]
 pub(crate) struct LoginToken {
     pub token: String,
     expires_in: Option<i64>,
     pub creation_time: Option<chrono::DateTime<chrono::offset::FixedOffset>>,
 }
 
+impl std::fmt::Debug for LoginToken {
+    fn fmt(&self, f: &mut std::fmt::Formatter<'_>) -> std::fmt::Result {
+        f.debug_struct("LoginToken")
+            .field("expires_in", &self.expires_in)
+            .field("creation_time", &self.creation_time)
+            .finish()
+    }
+}
+
+#[derive(Debug)]
 struct ClientState {
     login_token: LoginToken,
 }
@@ -216,12 +236,12 @@ impl ClientBuilder {
         let role_id = approle_id.unwrap_or(&secret_id);
 
         let base_url = url::Url::parse(host)?;
-        let url = base_url.join("v1/approle_login")?;
+        let url = base_url.join("v1/approle-login")?;
         let data = json!({
-            "role_id": role_id,
-            "secret_id": secret_id,
+            "role-id": role_id,
+            "secret-id": secret_id,
         });
-        Client::do_login(&self, base_url, url, data).await
+        Client::do_login(self, base_url, url, data).await
     }
 
     /// Login to an avassa Control Tower or Edge Enforcer instance. If possible,
@@ -235,7 +255,13 @@ impl ClientBuilder {
             "username":username,
             "password":password
         });
-        Client::do_login(&self, base_url, url, data).await
+        Client::do_login(self, base_url, url, data).await
+    }
+}
+
+impl Default for ClientBuilder {
+    fn default() -> Self {
+        Self::new()
     }
 }
 
@@ -252,6 +278,19 @@ pub struct Client {
     disable_cert_verification: bool,
 }
 
+impl std::fmt::Debug for Client {
+    fn fmt(&self, f: &mut std::fmt::Formatter<'_>) -> std::fmt::Result {
+        f.debug_struct("Client")
+            .field("base_url", &self.base_url)
+            .field("websocket_url", &self.websocket_url)
+            .field("state", &self.state)
+            .field("client", &self.client)
+            .field("disable_hostname_check", &self.disable_hostname_check)
+            .field("disable_cert_verification", &self.disable_cert_verification)
+            .finish()
+    }
+}
+
 impl Client {
     /// Create a Client builder
     pub fn builder() -> ClientBuilder {
@@ -289,7 +328,7 @@ impl Client {
             Self::new(builder, client, base_url, login_token)
         } else {
             let text = result.text().await?;
-            debug!("login returned {}", text);
+            tracing::debug!("login returned {}", text);
             Err(Error::LoginFailure(text))
         }
     }
@@ -300,7 +339,7 @@ impl Client {
         base_url: url::Url,
         login_token: LoginToken,
     ) -> Result<Self> {
-        let websocket_url = url::Url::parse(&format!("ws://{}/v1/ws/", base_url.host_port()?))?;
+        let websocket_url = url::Url::parse(&format!("wss://{}/v1/ws/", base_url.host_port()?))?;
 
         let state = ClientState { login_token };
 
@@ -343,11 +382,9 @@ impl Client {
         let result = builder.send().await?;
 
         if result.status().is_success() {
-            let text = result.text().await?;
-            let res = serde_json::from_str(&text)?;
+            let res = result.json().await?;
             Ok(res)
         } else {
-            error!("HTTP call failed");
             Err(Error::WebServer(
                 result.status().as_u16(),
                 result.status().to_string(),
@@ -364,7 +401,7 @@ impl Client {
         let url = self.base_url.join(path)?;
         let token = self.state.lock().await.login_token.token.clone();
 
-        debug!("POST {} {:?}", url, data);
+        tracing::debug!("POST {} {:?}", url, data);
 
         let result = self
             .client
@@ -374,6 +411,54 @@ impl Client {
             .send()
             .await?;
 
+        if result.status().is_success() {
+            let resp = result.bytes().await?;
+
+            let mut responses: Vec<serde_json::Value> = Vec::new();
+            let decoder = serde_json::Deserializer::from_slice(&resp);
+
+            for v in decoder.into_iter() {
+                responses.push(v?);
+            }
+
+            if responses.len() > 1 {
+                // Convert to a JSON array
+                Ok(serde_json::Value::Array(responses))
+            } else if responses.len() == 1 {
+                Ok(responses.into_iter().next().unwrap())
+            } else {
+                Ok(serde_json::Value::Object(Default::default()))
+            }
+        } else {
+            tracing::error!("POST call failed");
+            let status = result.status();
+            let resp = result.json().await;
+            match resp {
+                Ok(resp) => Err(Error::REST(resp)),
+                Err(_) => Err(Error::WebServer(status.as_u16(), status.to_string())),
+            }
+        }
+    }
+
+    /// PUT arbitrary JSON to a path
+    pub async fn put_json(
+        &self,
+        path: &str,
+        data: &serde_json::Value,
+    ) -> Result<serde_json::Value> {
+        let url = self.base_url.join(path)?;
+        let token = self.state.lock().await.login_token.token.clone();
+
+        tracing::debug!("PUT {} {:?}", url, data);
+
+        let result = self
+            .client
+            .put(url)
+            .json(&data)
+            .bearer_auth(&token)
+            .send()
+            .await?;
+
         if result.status().is_success() {
             use std::error::Error;
             let resp = result.json().await.or_else(|e| match e {
@@ -393,7 +478,7 @@ impl Client {
             })?;
             Ok(resp)
         } else {
-            error!("POST call failed");
+            tracing::error!("PUT call failed");
             let status = result.status();
             let resp = result.json().await;
             match resp {
@@ -403,6 +488,11 @@ impl Client {
         }
     }
 
+    /// Remove me
+    pub fn get_client(&self) -> reqwest::Client {
+        self.client.clone()
+    }
+
     /// Open a volga producer on a topic
     pub async fn volga_open_producer(
         &self,
@@ -410,21 +500,21 @@ impl Client {
         topic: &str,
         options: volga::Options,
     ) -> Result<volga::Producer> {
-        crate::volga::ProducerBuilder::new(&self, producer_name, topic, options)?
+        crate::volga::ProducerBuilder::new(self, producer_name, topic, options)?
             .set_options(options)
             .connect()
             .await
     }
 
-    /// Open a volga NAT producer on a topic in a uDC
+    /// Open a volga NAT producer on a topic in a site
     pub async fn volga_open_nat_producer(
         &self,
         producer_name: &str,
         topic: &str,
-        udc: &str,
+        site: &str,
         options: volga::Options,
     ) -> Result<volga::Producer> {
-        crate::volga::ProducerBuilder::new_nat(&self, producer_name, topic, udc, options)?
+        crate::volga::ProducerBuilder::new_nat(self, producer_name, topic, site, options)?
             .set_options(options)
             .connect()
             .await
@@ -437,7 +527,7 @@ impl Client {
         topic: &str,
         options: crate::volga::ConsumerOptions,
     ) -> Result<volga::Consumer> {
-        crate::volga::ConsumerBuilder::new(&self, consumer_name, topic)?
+        crate::volga::ConsumerBuilder::new(self, consumer_name, topic)?
             .set_options(options)
             .connect()
             .await
@@ -448,10 +538,10 @@ impl Client {
         &self,
         consumer_name: &str,
         topic: &str,
-        udc: &str,
+        site: &str,
         options: crate::volga::ConsumerOptions,
     ) -> Result<volga::Consumer> {
-        crate::volga::ConsumerBuilder::new_nat(&self, consumer_name, topic, udc)?
+        crate::volga::ConsumerBuilder::new_nat(self, consumer_name, topic, site)?
             .set_options(options)
             .connect()
             .await
@@ -480,6 +570,11 @@ impl Client {
     pub async fn volga_open_log_query(&self, query: &volga::Query) -> Result<volga::QueryStream> {
         volga::QueryStream::new(self, query).await
     }
+
+    /// Try to open a Strongbox Vault
+    pub async fn open_strongbox_vault(&self, vault: &str) -> Result<strongbox::Vault> {
+        strongbox::Vault::open(self, vault).await
+    }
 }
 
 pub(crate) trait URLExt {
